/*
 *	NEWS-0.4.6 - GTK-based RSS Ticker - Copyright (C) Emmanuel Thomas-Maurin 2009-2011
 *	<manutm007@gmail.com>
 *
 * 	This program is free software: you can redistribute it and/or modify
 * 	it under the terms of the GNU General Public License as published by
 * 	the Free Software Foundation, either version 3 of the License, or
 * 	(at your option) any later version.
 *
 * 	This program is distributed in the hope that it will be useful,
 * 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 * 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * 	GNU General Public License for more details.
 *
 * 	You should have received a copy of the GNU General Public License
 * 	along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "news.h"

static GrEnv		*env;
static Resource		*resrc;
static Params		*prm;
static int		shift_counter = 0;
static int		rss_ttl;
#if USE_GUI
static GtkWidget	*menu_bar;
/*static int		menu_included;*/
static GtkWidget	*popup_menu;
static GtkAccelGroup	*popup_menu_accel_group;
static int		popup_menu_accel_group_attached;
static int		position_x_in_drwa;
#endif
static GtkWidget	*main_box;	/* window layout -> vbox or hbox */
static GtkWidget	*hbox2, *vbox_ticker, *vbox_clock;
static int		speed_up_flag = FALSE;
static int		speed_down_flag = FALSE;
static int		instance_id = 0;

#ifdef G_OS_WIN32
FILE			*stdout_fp, *stderr_fp;
#endif
extern const char	*help_str1[];
extern const char	*license_str1[];
extern const char	*license_str2;

/* prototypes for some static funcs */
static gint shift2left_callback();

#if USE_GUI
void toogle_speed_up_flag()
{
	speed_up_flag = TRUE;
	return;
}

void toogle_speed_down_flag()
{
	speed_down_flag = TRUE;
	return;
}
#endif

void check_main_win_always_on_top()
{
	if (prm->always_on_top == 'y')
		gtk_window_set_keep_above(GTK_WINDOW(get_gr_env()->win), TRUE);
	else
		gtk_window_set_keep_above(GTK_WINDOW(get_gr_env()->win), FALSE);
	return;
}

#if USE_GUI
/*
 * funct_name2(***NO args***) only used to call funct_name(***args***)
 */
static void open_new_txt_file2()
{
	open_new_txt_file(resrc, &rss_ttl);
	return;
}

static void get_new_url2()
{
	get_new_url(resrc, prm, &rss_ttl);
	return;
}

static void modify_params2()
{
	if (modify_params(prm) == 1)	/* 1 means we want to reopen the dialog */
		modify_params2(prm);
	return;
}

#ifdef G_OS_WIN32
static int easily_link_with_browser2()
{
	char		browser_cmd[512], tmp[1024];
	unsigned int	i, count, start = 0, end = 0;
	int		exit_status = NO;

	if (get_default_browser_from_win32registry(browser_cmd) == WIN32REGKEY_FOUND) {
		/* find 1st str in browser_cmd */
		str_n_cpy(tmp, browser_cmd, 511);
		for (i = 0, count = 0; i < strlen(tmp); i++) {
			if (tmp[i] == '"') {
				count++;
				if (count == 1)
					start = i + 1;
				else if (count == 2)
					end = i - 1;
			}
		}
		str_n_cpy(browser_cmd, tmp + start, MIN((end - start + 1), 511));
		snprintf(tmp, 1024,
			"\nThis is the Shell command that open your default Browser:\n\n%s\n\n"
			"Do you want to use it (recommended) ?\n", browser_cmd);
		if (question_win(tmp) == YES) {
			exit_status = YES;
			str_n_cpy(prm->open_link_cmd, browser_cmd, FILENAME_MAXLEN);
			str_n_cpy(prm->open_link_args, "", FILENAME_MAXLEN);
			if (question_win("\nSave this parameter ?\n") == YES)
				save_to_config_file(prm);
		} else
			info_win("\nCancelled.\n"
				"You may too set this parameter manually in the Preferences window.\n");
	} else
		info_win("\nCan't find Browser shell command.\n"
			"You will have to set this parameter manually in the Preferences window.\n");
	return exit_status;
}

static void easily_link_with_browser()
{
	gtk_window_set_keep_above(GTK_WINDOW(env->win), FALSE);
	easily_link_with_browser2();
	check_main_win_always_on_top();
	return;
}
#endif

static void ticker_play()
{
	env->suspend_rq = FALSE;
	return;
}

static void ticker_pause()
{
	env->suspend_rq = TRUE;
	return;
}

static void ticker_reload()
{
	env->reload_rq = TRUE;
	env->suspend_rq = FALSE;
	return;
}

static int get_visible_link()
{
	int	i, location_on_pixmap;

	location_on_pixmap = shift_counter * prm->shift_size + position_x_in_drwa;
	for (i = 0; i < NFEEDLINKANDOFFSETMAX; i++) {
		if (resrc->link_and_offset[i].offset_in_pixmap > location_on_pixmap) {
			str_n_cpy(env->active_link, resrc->link_and_offset[i].url, FILENAME_MAXLEN);
			return TRUE;
		}
	}
	env->active_link[0] = '\0';
	return FALSE;
}

static void open_link()
{
	char		tmp1[FILENAME_MAXLEN + 1], tmp2[FILENAME_MAXLEN + 1];
#ifdef G_OS_WIN32
	char		*tmp;
#endif
	char		*argv[32];	/* up to 32 - 3 (prog name, url, NULL) args */
	GPid		pid;
	GError		*error = NULL;
	int		i, j;

	if (prm->open_link_cmd[0] == '\0') {
#ifndef G_OS_WIN32
		/* TODO: should look for default browser */
		warning("Can't launch Browser: no command is defined.\n",
			"Please set the 'Open in Browser' option in the Preferences window.", "", "");
		return;
#else
		easily_link_with_browser();
		if (prm->open_link_cmd[0] == '\0') {
			warning("Can't launch Browser: no command is defined.\n",
				"Please set the 'Open in Browser' option in the Preferences window.", "", "");
			return;
		}
#endif
	}

	if (env->active_link [0] == '\0') {
		warning("No link found", "", "", "");
		return;
	}

	fprintf(STD_OUT, "Spawning: %s %s %s\n", prm->open_link_cmd, prm->open_link_args, env->active_link);
#ifndef G_OS_WIN32
	str_n_cpy(tmp1, prm->open_link_cmd, FILENAME_MAXLEN);
#else
	tmp = g_win32_locale_filename_from_utf8(prm->open_link_cmd);
	str_n_cpy(tmp1, tmp, FILENAME_MAXLEN);
	g_free(tmp);
#endif
	str_n_cpy(tmp2, prm->open_link_args, FILENAME_MAXLEN);
	argv[0] = tmp1;
	for (i = 0, j = 1; tmp2[i] != '\0' && j < 32 + 1 - 3; j++) {
		argv[j] = &tmp2[i];
		while (tmp2[++i] != ' ')
			if (tmp2[i] == '\0')
				break;
		tmp2[i] = '\0';
		while (tmp2[++i] == ' ');
	}
	argv[j] = env->active_link;
	argv[j + 1] = NULL;
	if (!g_spawn_async(NULL, argv, NULL, G_SPAWN_SEARCH_PATH, NULL, NULL, &pid, &error)) {
		warning("News: Can't create process ", argv[0], " - ", error->message);
#ifdef G_OS_WIN32
		if (easily_link_with_browser2() == YES)
			open_link();
#endif
	} else
		g_spawn_close_pid(pid);
	return;
}

static void open_link2()
{
	position_x_in_drwa = env->drwa_width / 2;
	get_visible_link();
	open_link();
	return;
}

static gint left_click_on_drawing_area(GtkWidget *widget, GdkEventButton *event, gpointer unused)
{
	widget = widget;
	unused = unused;
	if (event->type == GDK_BUTTON_PRESS && event->button == 1) {	/* 1 = mouse left button */
		/* get pointer x location on drawing_area */
		position_x_in_drwa = event->x;
		get_visible_link();
		open_link();
		return TRUE;
	} else
		return FALSE;
}

static gint right_click_on_drawing_area(GtkWidget *widget, GdkEventButton *event, gpointer unused)
{
	widget = widget;
	unused = unused;
	if (event->type == GDK_BUTTON_PRESS && event->button == 3) {	/* 3 = mouse right button */
		/* popup main menu */
		gtk_menu_popup(GTK_MENU(popup_menu), NULL, NULL, NULL, NULL, event->button, event->time);
		return TRUE;
	} else
		return FALSE;
}

/*
 * menu stuff
 */
# if MENU_V
static GtkItemFactoryEntry menu_item[] = {
	{"/_"MENU_LAB,			NULL, NULL, 0, "<Branch>", NULL},

	{"/"MENU_LAB"/File",		NULL, NULL, 0, "<Branch>",
					(gconstpointer)GTK_STOCK_FILE},

	{"/"MENU_LAB"/File/_Open Text File", "<control>O", open_new_txt_file2, 0, "<StockItem>",
					(gconstpointer)GTK_STOCK_OPEN},

	{"/"MENU_LAB"/File/Open _RSS Feed", "<control>R", get_new_url2, 0, "<StockItem>",
					(gconstpointer)"news_rss_icon"},

	{"/"MENU_LAB"/File/sep",	NULL, NULL, 0, "<Separator>", NULL},

	{"/"MENU_LAB"/File/_Quit",	"<control>Q", gtk_main_quit, 0, "<StockItem>",
					(gconstpointer)GTK_STOCK_QUIT},

	{"/"MENU_LAB"/Edit",		NULL, NULL, 0, "<Branch>", NULL},

	{"/"MENU_LAB"/Edit/_Preferences","<control>P", modify_params2, 0, "<StockItem>",
					(gconstpointer)GTK_STOCK_PREFERENCES},
	{"/"MENU_LAB"/Control",		NULL, NULL, 0, "<Branch>", NULL},

	{"/"MENU_LAB"/Control/Open Link in _Browser", "<control>B", open_link2, 0, "<StockItem>",
					(gconstpointer)GTK_STOCK_JUMP_TO},

	{"/"MENU_LAB"/Control/Play",	"<control>J", ticker_play, 0, "<StockItem>",
					(gconstpointer)GTK_STOCK_MEDIA_PLAY},

	{"/"MENU_LAB"/Control/Pause",	"<control>K", ticker_pause, 0, "<StockItem>",
					(gconstpointer)GTK_STOCK_MEDIA_PAUSE},

	{"/"MENU_LAB"/Control/Reload",	"<control>L", ticker_reload, 0, "<StockItem>",
					(gconstpointer)GTK_STOCK_MEDIA_REWIND},

	{"/"MENU_LAB"/Control/Speed Up", "<control>U", toogle_speed_up_flag, 0, NULL, NULL},

	{"/"MENU_LAB"/Control/Speed Down", "<control>D", toogle_speed_down_flag, 0, NULL, NULL},

	{"/"MENU_LAB"/Help",		NULL, NULL, 0, "<Branch>", NULL},

	{"/"MENU_LAB"/Help/Quick Help",	"F1", help_win, 0, "<StockItem>",
					(gconstpointer) GTK_STOCK_HELP},

	{"/"MENU_LAB"/Help/Online Help", "<control>H", online_help, 0, NULL, NULL},

	{"/"MENU_LAB"/Help/_About",	"<control>A", about_win, 0, "<StockItem>",
					(gconstpointer)GTK_STOCK_ABOUT},
};
# else
static GtkItemFactoryEntry menu_item[] = {
	{"/_File",		NULL, NULL, 0, "<Branch>", NULL},

	{"/File/_Open Text File", "<control>O", open_new_txt_file2, 0, "<StockItem>",
				(gconstpointer)GTK_STOCK_OPEN},

	{"/File/Open _RSS Feed", "<control>R", get_new_url2, 0, "<StockItem>",
				(gconstpointer)"news_rss_icon"},

	{"/File/sep",		NULL, NULL, 0, "<Separator>", NULL},

	{"/File/_Quit",		"<control>Q", gtk_main_quit, 0, "<StockItem>",
				(gconstpointer)GTK_STOCK_QUIT},

	{"/_Edit",		NULL, NULL, 0, "<Branch>", NULL},

	{"/Edit/_Preferences",	"<control>P", modify_params2, 0, "<StockItem>",
				(gconstpointer)GTK_STOCK_PREFERENCES},
	{"/_Control",		NULL, NULL, 0, "<Branch>", NULL},

	{"/Control/Open Link in _Browser", "<control>B", open_link2, 0, "<StockItem>",
				(gconstpointer)GTK_STOCK_JUMP_TO},

	{"/Control/Play",	"<control>J", ticker_play, 0, "<StockItem>",
				(gconstpointer)GTK_STOCK_MEDIA_PLAY},

	{"/Control/Pause",	"<control>K", ticker_pause, 0, "<StockItem>",
				(gconstpointer)GTK_STOCK_MEDIA_PAUSE},

	{"/Control/Reload",	"<control>L", ticker_reload, 0, "<StockItem>",
				(gconstpointer)GTK_STOCK_MEDIA_REWIND},

	{"/Control/Speed Up",	"<control>U", toogle_speed_up_flag, 0, NULL, NULL},

	{"/Control/Speed Down",	"<control>D", toogle_speed_down_flag, 0, NULL, NULL},

	{"/_Help",		NULL, NULL, 0, "<Branch>", NULL},

	{"/Help/Quick Help",	"F1", help_win, 0, "<StockItem>",
				(gconstpointer)GTK_STOCK_HELP},

	{"/Help/Online Help",	"<control>H", online_help, 0, NULL, NULL},

	{"/Help/_About",	"<control>A", about_win, 0, "<StockItem>",
				(gconstpointer)GTK_STOCK_ABOUT},
};
# endif

static GtkItemFactoryEntry popup_menu_item[] = {
	{"/_File",		NULL, NULL, 0, "<Branch>", NULL},

	{"/File/_Open Text File", "<control>O", open_new_txt_file2, 0, "<StockItem>",
				(gconstpointer)GTK_STOCK_OPEN},

	{"/File/Open _RSS Feed", "<control>R", get_new_url2, 0, "<StockItem>",
				(gconstpointer)"news_rss_icon"},

	{"/File/sep",		NULL, NULL, 0, "<Separator>", NULL},

	{"/File/_Quit",		"<control>Q", gtk_main_quit, 0, "<StockItem>",
				(gconstpointer)GTK_STOCK_QUIT},

	{"/_Edit",		NULL, NULL, 0, "<Branch>", NULL},

	{"/Edit/_Preferences",	"<control>P", modify_params2, 0, "<StockItem>",
				(gconstpointer)GTK_STOCK_PREFERENCES},
	{"/_Control",		NULL, NULL, 0, "<Branch>", NULL},

	{"/Control/Open Link in _Browser", "<control>B", open_link2, 0, "<StockItem>",
				(gconstpointer)GTK_STOCK_JUMP_TO},

	{"/Control/Play",	"<control>J", ticker_play, 0, "<StockItem>",
				(gconstpointer)GTK_STOCK_MEDIA_PLAY},

	{"/Control/Pause",	"<control>K", ticker_pause, 0, "<StockItem>",
				(gconstpointer)GTK_STOCK_MEDIA_PAUSE},

	{"/Control/Reload",	"<control>L", ticker_reload, 0, "<StockItem>",
				(gconstpointer)GTK_STOCK_MEDIA_REWIND},

	{"/Control/Speed Up",	"<control>U", toogle_speed_up_flag, 0, NULL, NULL},

	{"/Control/Speed Down",	"<control>D", toogle_speed_down_flag, 0, NULL, NULL},

	{"/_Help",		NULL, NULL, 0, "<Branch>", NULL},

	{"/Help/Quick Help",	"F1", help_win, 0, "<StockItem>",
				(gconstpointer)GTK_STOCK_HELP},

	{"/Help/Online Help",	"<control>H", online_help, 0, NULL, NULL},

	{"/Help/_About",	"<control>A", about_win, 0, "<StockItem>",
				(gconstpointer)GTK_STOCK_ABOUT},
};

static gint n_menu_items = sizeof(menu_item) / sizeof(menu_item[0]);

static gint n_popup_menu_items = sizeof(popup_menu_item) / sizeof(popup_menu_item[0]);

static GtkWidget *get_menubar_menu(GtkWidget *win)
{
	GtkItemFactory	*item_factory;
	GtkAccelGroup	*menu_bar_accel_group;

	menu_bar_accel_group = gtk_accel_group_new();
	item_factory = gtk_item_factory_new(GTK_TYPE_MENU_BAR, "<main>", menu_bar_accel_group);
	gtk_item_factory_create_items(item_factory, n_menu_items, menu_item, NULL);
	gtk_window_add_accel_group(GTK_WINDOW(win), menu_bar_accel_group);
	return gtk_item_factory_get_widget(item_factory, "<main>");
}

static GtkWidget *get_popup_menu_menu(GtkWidget *win)
{
	GtkItemFactory	*item_factory;
	/* set as global
	GtkAccelGroup	*popup_menu_accel_group;*/

	win = win;
	popup_menu_accel_group = gtk_accel_group_new();
	item_factory = gtk_item_factory_new(GTK_TYPE_MENU, "<main>", popup_menu_accel_group);
	gtk_item_factory_create_items(item_factory, n_popup_menu_items, popup_menu_item, NULL);
	return gtk_item_factory_get_widget(item_factory, "<main>");
}
#endif

static gint update_win_dims()
{
	GtkRequisition	req;
	gint		v_menu_width;

	if (USE_GUI && MENU_V && prm->windec == 'y') {
		gtk_widget_size_request(menu_bar, &req);
		v_menu_width = req.width;
	} else
		v_menu_width = 0;

	gtk_widget_set_size_request(env->drw_a, env->drwa_width, env->drwa_height);
	gtk_widget_set_size_request(env->drwa_clock, get_clock_width(prm), env->drwa_height);

	if (prm->windec == 'y') {
		if(MENU_V)
			gtk_widget_set_size_request(env->win, 0, 0);
		else
			gtk_widget_set_size_request(env->win, -1, -1);
	} else
		gtk_widget_set_size_request(env->win, 0, 0);

	gtk_window_resize(GTK_WINDOW(env->win),
		v_menu_width + env->drwa_width + get_clock_width(prm), env->drwa_height);

	gtk_window_unmaximize(GTK_WINDOW(env->win));
	/*
	 * not about win dims but this must be checked too
	 * is there another place for this check?
	 */
	check_main_win_always_on_top();
	return TRUE;
}

/*
 * create new or next pixmap & compute stuff to redraw window
 * *** TODO: add extra parameter to read file from start ***
 */
static int compute_pixmap_and_win()
{
#if (USE_GUI && MENU_V)
	GtkRequisition	req;
#endif
	char 	fname[FONT_NAME_MAXLEN + 1], fsize[FONT_SIZE_MAXLEN + 1];
	char	cfname[FONT_NAME_MAXLEN + 1], cfsize[FONT_SIZE_MAXLEN + 1];
	gint	height1, height2;
	int	size, render_exit_status, i;
	char	tmp[80];

	/*
	 * compute font size from requested height if > 0
	 */
	split_font(prm->font_n_s, fname, fsize);
	if (prm->win_h > 0)
		snprintf(fsize, FONT_SIZE_MAXLEN + 1, "%3d",
			get_fsize_from_layout_height((gint)prm->win_h, fname));
	/* in all cases, font size can't be > FONT_MAXSIZE */
	if (atoi(fsize) > FONT_MAXSIZE)
		snprintf(fsize, FONT_SIZE_MAXLEN + 1, "%3d", FONT_MAXSIZE);
	compact_font(prm->font_n_s, fname, fsize);
	/*
	 * compute clock font size
	 * clock height = ticker height unless clock font size is set
	 * in which case clock height is always <= ticker height
	 */
	split_font(prm->clock_font_n_s, cfname, cfsize);
	if (strcmp(fname, cfname) == 0) {
		if (atoi(cfsize) > atoi(fsize))
			str_n_cpy(cfsize, fsize, FONT_SIZE_MAXLEN);
	} else {
		height1 = get_layout_height_from_fnamesize(prm->clock_font_n_s);
		height2 = get_layout_height_from_fnamesize(prm->font_n_s);
		if (height1 > height2) {
			size = get_fsize_from_layout_height(height2, cfname);
			snprintf(cfsize, FONT_SIZE_MAXLEN + 1, "%3d", size);
		}
	}
	compact_font(prm->clock_font_n_s, cfname, cfsize);
	/*
	 * compute ticker width
	 */
	if (prm->win_w >= DRWA_WIDTH_MIN && prm->win_w <= env->screen_w)
		env->drwa_width = prm->win_w;
	else if (prm->win_w < DRWA_WIDTH_MIN)
		env->drwa_width = DRWA_WIDTH_MIN;
	else if (prm->win_w > env->screen_w)
		env->drwa_width = env->screen_w;

#if (USE_GUI && MENU_V)
	if (prm->windec == 'y') {
		gtk_widget_size_request(menu_bar, &req);
		env->drwa_width -= req.width;
	}
#endif
	env->drwa_width -= get_clock_width(prm);
	/*
	 * create pixmap of rendered text (one long single line)
	 */
	if(resrc->id[0] == '\0')
		if (resrc->fp != NULL) {
			fclose(resrc->fp);
			resrc->fp = NULL;

			for (i = 0; i < NFEEDLINKANDOFFSETMAX; i++) {
				resrc->link_and_offset[i].url[0] = '\0';
				resrc->link_and_offset[i].offset_in_pixmap = 0;
			}

		}

	env->pixmap = render_stream_to_pixmap(resrc->fp, resrc->link_and_offset, prm, &render_exit_status);

	if (render_exit_status != RENDER_OK && render_exit_status != RENDER_NO_RESOURCE) {
		switch(render_exit_status) {
			case RENDER_NULL_ENDINGSTR:
				warning("render_stream_to_pixmap():",
					"ending string = NULL", "", "");
				break;
			case RENDER_GETLINE_ERROR:
				warning("render_stream_to_pixmap():",
					"getline() error", "", "");
				break;
			case RENDER_LINE_TOO_LONG:
				warning("render_stream_to_pixmap():",
					"line too long\n", "", "");
				break;
			case RENDER_PIXMAP_TOO_WIDE:
				big_error(RENDER_PIXMAP_TOO_WIDE,
					"render_stream_to_pixmap():",
					"pixmap too wide (> 32 K pixels)", "", "");
				break;
			default:
				warning("render_stream_to_pixmap():",
					"unknown error", "", "");
				break;
		}
	}
	if (env->pixmap == NULL)
		big_error(render_exit_status,
			"render_stream_to_pixmap():", "pixmap = NULL", "", "");

	gdk_drawable_get_size(env->pixmap, &env->pixm_width, &env->pixm_height);
	if (env->pixm_height  > env->screen_h)
		env->drwa_height = env->screen_h;
	else
		env->drwa_height = env->pixm_height;
	/*
	 * window layout stuff
	 */
	/* win title */
	if (resrc->resrc_type == RESRC_URL)
		snprintf(tmp, 80, "News-%s  |  %s", APP_VERSION_NUMBER, resrc->feed_title);
	else if (resrc->resrc_type == RESRC_FILE)
		snprintf(tmp, 80, "News-%s  |  %s", APP_VERSION_NUMBER, resrc->id);
	gtk_window_set_title(GTK_WINDOW(env->win), tmp);

	/* move win */
	gtk_window_move(GTK_WINDOW(env->win), prm->win_x, prm->win_y);

	/* win decoration */
	if (prm->windec == 'y') {
		gtk_window_set_decorated(GTK_WINDOW(env->win), TRUE);
		if (!popup_menu_accel_group_attached) {
			gtk_window_add_accel_group(GTK_WINDOW(env->win), popup_menu_accel_group);
			popup_menu_accel_group_attached = TRUE;
		}
	} else {
		gtk_window_set_decorated(GTK_WINDOW(env->win), FALSE);
		if (popup_menu_accel_group_attached) {
			gtk_window_remove_accel_group(GTK_WINDOW(env->win), popup_menu_accel_group);
			popup_menu_accel_group_attached = FALSE;
		}
	}

	/* ### deprecated ###
	 * adding menu_bar to main_box if windec='y' - is this necessary? anyways, it doesn't work
	 *
	if (prm->windec == 'y') {
		if (!menu_included) {
			gtk_container_add(GTK_CONTAINER(main_box), menu_bar);
			gtk_box_reorder_child(GTK_BOX(main_box), menu_bar, 0);
			menu_included = TRUE;
		}
	} else {
		if (menu_included) {
			g_object_ref(menu_bar);
			gtk_container_remove(GTK_CONTAINER(main_box), menu_bar);
			menu_included = FALSE;
		}
	}*/

	/* clock */
	if (prm->clock == 'l')
		gtk_box_reorder_child(GTK_BOX(hbox2), vbox_clock, 0);
	else if (prm->clock == 'r')
		gtk_box_reorder_child(GTK_BOX(hbox2), vbox_clock, 1);

	/* whole window transparency - what about ticker only? */
	gtk_window_set_opacity(GTK_WINDOW(env->win), prm->win_transparency);

	return render_exit_status;
}

/*
 * timeout handler to render one part of pixmap onto drawing area
 * image is shifted to the left by <shift_size> pixels
 *
 * listen to suspend_rq, reload_rq and compute_rq
 */
static gint shift2left_callback()
{
	if (env->suspend_rq)
		return TRUE;	/* does nothing - just return */
	else if ((shift_counter * prm->shift_size < env->pixm_width - env->drwa_width - 8) \
			&& !env->reload_rq && !env->compute_rq \
			&& GDK_IS_DRAWABLE((env->drw_a)->window) && GDK_IS_GC(env->gc)) {
		/* draw onto ticker area */
		gdk_draw_drawable(GDK_DRAWABLE((env->drw_a)->window),
			env->gc, env->pixmap, shift_counter++ * prm->shift_size, 0, 0, 0,
			env->drwa_width, env->drwa_height);

		if (speed_up_flag == TRUE || speed_down_flag == TRUE) {
			if (speed_up_flag == TRUE) {
				speed_up_flag = FALSE;
				if (prm-> delay > 1)
					prm->delay--;
				else
					return TRUE;
			} else if (speed_down_flag == TRUE) {
				speed_down_flag = FALSE;
				if (prm-> delay < 50)
					prm->delay++;
				else
					return TRUE;
			}
			g_timeout_add_full(G_PRIORITY_DEFAULT, prm->delay, shift2left_callback, NULL, NULL);
			return FALSE;
		} else
			return TRUE;
	} else {
		/*
		 * when pixmap scrolling completed or if request to reload or
		 * recompute everything, will create new or next one
		 */
		env->suspend_rq = TRUE;
		if (env->pixmap != NULL) {
			g_object_unref(env->pixmap);
			env->pixmap = NULL;
		}
		if (env->reload_rq) {
			load_resource(resrc, &rss_ttl);
			check_time_load_resource(TIME_RESET);
		}
		compute_pixmap_and_win();
		shift_counter = prm->shift_size;	/* to avoid displaying twice the same thing */
		g_timeout_add_full(G_PRIORITY_DEFAULT, prm->delay, shift2left_callback, NULL, NULL);
		env->reload_rq = FALSE;
		env->compute_rq = FALSE;
		env->suspend_rq = FALSE;
		return FALSE;
	}
}

static gint display_time2()
{
	display_time(prm);
	return TRUE;
}

static gint get_win_position()	/* only if win is 'draggable' */
{
	if (prm->windec == 'y')
		gtk_window_get_position(GTK_WINDOW(env->win), &prm->win_x, &prm->win_y);
	return TRUE;
}

/*
 * reload stuff every <rss_ttl * 60> seconds
 */
/*static gint load_resource2()
{
	env->suspend_rq = TRUE;
	load_resource(resrc, &rss_ttl);
	compute_pixmap_and_win();
	g_timeout_add_seconds(rss_ttl * 60, load_resource2, NULL);
	env->suspend_rq = FALSE;
	return FALSE;
}*/
void check_time_load_resource(check_time_mode mode)
{
	static unsigned long	elapsed_time_in_sec;

	if (mode == TIME_RESET)
		elapsed_time_in_sec = 0;
	else if (mode == TIME_CHECK) {
		if (resrc->resrc_type == RESRC_FILE)
			rss_ttl = prm->rss_refresh;
		if ((++elapsed_time_in_sec) / 60 >= (unsigned long)rss_ttl) {
			elapsed_time_in_sec = 0;
			env->reload_rq = TRUE;
		}
	}
}

gint check_time_load_resource2()
{
	check_time_load_resource(TIME_CHECK);
	return TRUE;
}

GrEnv *get_gr_env()
{
	return env;
}

Resource *get_resource()
{
	return resrc;
}

Params *get_params()
{
	return prm;
}

static void quit_with_cli_info(news_error_code error_code)
{
	if (resrc->fp != NULL)
		fclose(resrc->fp);
	xmlCleanupParser();
	free2(env);
	free2(resrc);
	free2(prm);
	fprintf(STD_ERR, "Try 'news --help' for more information\n");
	exit(error_code);
}

static void print_help()
{
	int	i;

	for (i = 0; help_str1[i] != NULL; i++)
		fprintf(STD_OUT, "%s", help_str1[i]);
	fprintf(STD_OUT, "\n");
	return;
}

static void print_license()
{
	int	i;

	for (i = 0; license_str1[i] != NULL; i++)
		fprintf(STD_OUT, "%s", license_str1[i]);
	fprintf(STD_OUT, "%s\n", license_str2);
	return;
}

int get_instance_id()
{
	return instance_id;
}

/* reload resource as soon as scrolling has started to update all display params
 * TODO: reset open stream to begining and recompute instead of reload (which
 * is useless) -> must 1st make sure this doesn't break anything
 */
gint update_everything()
{
	if (shift_counter > 4) {
		env->reload_rq = TRUE;
		return FALSE;
	} else
		return TRUE;
}

int main(int argc, char *argv[])
{
	GdkVisual	*visual = NULL;
	GdkPixbuf	*pixb = NULL;
	GtkIconFactory	*icon_factory = NULL;
	GError		*error = NULL;
	FILE		*tmp_fp;
	int		n_options, n_resources, i;
	char		tmp[TMPSTR_SIZE + 1];

	gtk_init(&argc, &argv);

	env = malloc2(sizeof(GrEnv));
	resrc = malloc2(sizeof(Resource));
	prm = malloc2(sizeof(Params));

	env->win = env->drw_a = env->drwa_clock = NULL;
	env->pixmap = NULL;
	env->gc = NULL;
	env->colormap = NULL;
	env->screen_w = env->screen_h = env->depth = 0;
	env->drwa_width = env->drwa_height = env->pixm_width = env->pixm_height = 0;
	env->active_link[0] = '\0';
	env->suspend_rq = env->compute_rq = env->reload_rq = FALSE;

	/*
	 * toplevel window
	 */
	env->win = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_window_set_title(GTK_WINDOW(env->win), "News-"APP_VERSION_NUMBER);
	gtk_container_set_border_width(GTK_CONTAINER(env->win), 0);
	g_signal_connect(G_OBJECT(env->win), "delete_event", G_CALLBACK(gtk_main_quit), NULL);
	g_signal_connect(G_OBJECT(env->win), "destroy", G_CALLBACK(gtk_widget_destroy), &(env->win));

	LIBXML_TEST_VERSION
	xmlInitParser();

#ifdef G_OS_WIN32
	if (get_progfiles_dir() == NULL)
		big_error(WIN32V_ERROR, "Can't find Program Files directory", "", "", "");
	else if (get_appdata_dir() == NULL)
		big_error(WIN32V_ERROR, "Can't find Application Data directory", "", "", "");
#endif
	/*
	 * news dir in user home dir
	 */
#ifndef G_OS_WIN32
	str_n_cpy(tmp, usr_home_dir(), TMPSTR_SIZE - 64);
	str_n_cat(tmp, "/"NEWSPATH, 64);
#else
	str_n_cpy(tmp, get_appdata_dir(), TMPSTR_SIZE - 64);
	str_n_cat(tmp, "\\"NEWSPATH, 64);
#endif
	g_mkdir(tmp, S_IRWXU);
#ifdef G_OS_WIN32
	str_n_cpy(tmp, get_appdata_dir(), TMPSTR_SIZE - 64);
	str_n_cat(tmp, "\\"NEWSPATH"\\stdout", 64);
	if ((stdout_fp = g_fopen(tmp, "wb+")) == NULL)
		big_error(WIN32V_ERROR,
		"Can't create 'stdout' file in Application Data directory:", strerror(errno), "", "");
	str_n_cpy(tmp, get_appdata_dir(), TMPSTR_SIZE - 64);
	str_n_cat(tmp, "\\"NEWSPATH"\\stderr", 64);
	if ((stderr_fp = g_fopen(tmp, "wb+")) == NULL)
		big_error(WIN32V_ERROR,
		"Can't create 'stderr' file in Application Data directory", strerror(errno), "", "");
#endif
	/*
	 * get help, show license, set instance id or check available system fonts
	 */
	if (argc >= 2 && (strcmp(argv[1], "-help") == 0 || \
			strcmp(argv[1], "--help") == 0 || strcmp(argv[1], "-h") == 0)) {
		print_help();
		return 0;
	} else if (argc >= 2 && (strcmp(argv[1], "-license") == 0 || \
			strcmp(argv[1], "--license") == 0)) {
		print_license();
		return 0;
	} else if (argc >= 2 && strncmp(argv[1], "-instance-id=", strlen("-instance-id=")) == 0) {
		if ((instance_id = atoi(argv[1] + strlen("-instance-id="))) < 1 || instance_id > 99) {
			fprintf(STD_ERR, "Instance ID invalid value: %d\n", instance_id);
			instance_id = 0;
		}
	} else if (argc >= 2 && strcmp(argv[1], "-dumpfontlist") == 0) {
		dump_font_list();
		return 0;
	}

#ifndef G_OS_WIN32
	if ((tmp_fp = g_fopen(NEWS_ICON, "rb")) != NULL) {
		fclose(tmp_fp);
		gtk_window_set_icon_from_file(GTK_WINDOW(env->win), NEWS_ICON, NULL);
#else
	str_n_cpy(tmp, get_progfiles_dir(), TMPSTR_SIZE - 64);
	str_n_cat(tmp, "\\"NEWS_ICON, 64);
	if ((tmp_fp = g_fopen(tmp, "rb")) != NULL) {
		fclose(tmp_fp);
		gtk_window_set_icon_from_file(GTK_WINDOW(env->win), tmp, NULL);
#endif
	} else
		fprintf(STD_ERR, "Icon file not found\n");

	/*
	 * create rss icon
	 */
#ifndef G_OS_WIN32
	pixb = gdk_pixbuf_new_from_file(RSS_ICON, &error);
#else
	str_n_cpy(tmp, get_progfiles_dir(), TMPSTR_SIZE - 64);
	str_n_cat(tmp, "\\"RSS_ICON, 64);
	pixb = gdk_pixbuf_new_from_file(tmp, &error);
#endif
	icon_factory = gtk_icon_factory_new();
	gtk_icon_factory_add(icon_factory, "news_rss_icon", gtk_icon_set_new_from_pixbuf(pixb));
	gtk_icon_factory_add_default(icon_factory);
	if (pixb != NULL)
		g_object_unref(pixb);
	/*
	 * get some system graphical params
	 */
	visual = gdk_visual_get_system();
	env->depth = visual->depth;
	env->screen_w = gdk_screen_get_width(gtk_window_get_screen(GTK_WINDOW(env->win)));
	env->screen_h = gdk_screen_get_height(gtk_window_get_screen(GTK_WINDOW(env->win)));
	env->colormap = gdk_screen_get_rgba_colormap(gdk_screen_get_default());
	if (!GDK_IS_COLORMAP(env->colormap))
		env->colormap = gdk_colormap_new(gdk_rgb_get_visual(), TRUE);
	if (!GDK_IS_COLORMAP(env->colormap))
		warning("Can't get current colormap", "", "", "");

	resrc->id[0] = '\0';
	resrc->xmldump[0] = '\0';
	resrc->fp = NULL;
	for (i = 0; i < NFEEDLINKANDOFFSETMAX; i++) {
		resrc->link_and_offset[i].url[0] = '\0';
		resrc->link_and_offset[i].offset_in_pixmap = 0;
	}

	/*
	 * options and resource stuff
	 */
	set_default_options(prm);	/* init all params */
	get_config_file_options(prm);
	str_n_cpy(resrc->id, prm->homefeed, FILENAME_MAXLEN);

	if (argc > 1) {
		/*
		 * parse args: first -options... then resource - the rest is ignored
		 */
		n_options = 0;
		n_resources = 0;
		for (i = 1; i < argc; i++)
			if (argv[i][0] == '-')
				n_options ++;
			else
				break;
		if (i < argc)
			n_resources++;

		if (n_resources == 0) {
			fprintf(STD_ERR, "No resource specified\n");
			if (RESOURCE_NONE_QUIT)
				quit_with_cli_info(RESOURCE_NONE);
		} else
			/* open text file or get and open rss feed */
			str_n_cpy(resrc->id, argv[n_options + 1], FILENAME_MAXLEN);
		if (n_options > 0)
			if (parse_options_array(prm, n_options, (const char **)(argv + 1)) != OPTION_OK)
				if (OPTION_ERROR_QUIT)
					quit_with_cli_info(OPTION_ERROR);
	} else {
		fprintf(STD_ERR, "No resource specified\n");
		if (resrc->id[0] == '\0')  {
			if (RESOURCE_NONE_QUIT)
				quit_with_cli_info(RESOURCE_NONE);
		} else
			fprintf(STD_ERR, "Will use default resource (Homefeed): %s\n", resrc->id);
	}
	if (load_resource(resrc, &rss_ttl) != RESOURCE_OK)
		if (RESOURCE_NONE_QUIT)
			quit_with_cli_info(RESOURCE_NONE);
	rss_ttl = prm->rss_refresh;
	/*
	 * set window/menu layout
	 */
#if MENU_V
	main_box = gtk_hbox_new(FALSE, 0);
#else
	main_box = gtk_vbox_new(FALSE, 0);
#endif
	gtk_container_add(GTK_CONTAINER(env->win), main_box);
#if USE_GUI
	menu_bar = get_menubar_menu(env->win);
	/* ### deprecated ###
	 * adding menu_bar to main_box if windec='y' - is this necessary? anyways, it doesn't work
	 *
	g_object_ref(menu_bar);
	if (prm->windec == 'y') {
		gtk_box_pack_start(GTK_BOX(main_box), menu_bar, FALSE, FALSE, 0);
		menu_included = TRUE;
	} else
		menu_included = FALSE;*/
	popup_menu = get_popup_menu_menu(NULL);
	popup_menu_accel_group_attached = FALSE;
#endif
	hbox2 = gtk_hbox_new(FALSE, 0);
	gtk_box_pack_start(GTK_BOX(main_box), hbox2, FALSE, FALSE, 0);
	/*
	 * create 2 drawing areas
	 */
	env->drw_a = gtk_drawing_area_new();
	env->drwa_clock = gtk_drawing_area_new();
	gtk_widget_add_events(env->drw_a, GDK_ALL_EVENTS_MASK);	/* change mask? */
#if USE_GUI
	g_signal_connect(G_OBJECT(env->drw_a), "button-press-event", G_CALLBACK(left_click_on_drawing_area), NULL);
	g_signal_connect(G_OBJECT(env->drw_a), "button-press-event", G_CALLBACK(right_click_on_drawing_area), NULL);
#endif
	vbox_ticker = gtk_vbox_new(FALSE, 0);	/* vbox_ticker is only used to have a v-centered ticker */
	gtk_box_pack_start(GTK_BOX(vbox_ticker), GTK_WIDGET(env->drw_a), TRUE, FALSE, 0);
	gtk_box_pack_start(GTK_BOX(hbox2), vbox_ticker, FALSE, FALSE, 0);
	vbox_clock = gtk_vbox_new(FALSE, 0);	/* vbox_clock is only used to have a v-centered clock */
	gtk_box_pack_start(GTK_BOX(vbox_clock), GTK_WIDGET(env->drwa_clock), TRUE, FALSE, 0);
	gtk_box_pack_start(GTK_BOX(hbox2), vbox_clock, FALSE, FALSE, 0);
	/*
	 * create pixmap from text file or rss feed & compute stuff to draw window
	 */
	compute_pixmap_and_win();

	/* === timeout callbacks === */
	/*
	 * refresh the image every <delay> milliseconds
	 */
	g_timeout_add_full(G_PRIORITY_DEFAULT, prm->delay, shift2left_callback, NULL, NULL);
	/*
	 * display clock every 1 seconds
	 */
	g_timeout_add_full(G_PRIORITY_DEFAULT, 500, display_time2, NULL, NULL);
	/*
	 * get current win position (if 'dragged') every 0.5 seconds
	 */
	g_timeout_add_full(G_PRIORITY_DEFAULT, 500, get_win_position, NULL, NULL);
	/*
	 * check every sec until elapsed time in mn > <rss_ttl> mn then reload rss feed (or text file)
	 */
	check_time_load_resource(TIME_RESET);
	g_timeout_add_full(G_PRIORITY_DEFAULT, 1000, check_time_load_resource2, NULL, NULL);

	g_timeout_add_full(G_PRIORITY_DEFAULT, 500, update_win_dims, NULL, NULL);

	g_timeout_add_full(G_PRIORITY_DEFAULT, 100, update_everything, NULL, NULL);

	env->suspend_rq = FALSE;
	env->compute_rq = FALSE;
	env->reload_rq = TRUE;

	gtk_widget_show_all(env->win);

	update_win_dims();

	gtk_main();
	/*
	 * cleanup stuff
	 */
	if (resrc->fp != NULL)
		fclose(resrc->fp);
	xmlCleanupParser();
	free2(env);
	free2(resrc);
	free2(prm);
	return 0;
}
